;*****************************************************************************
;* LexOS 1.0 based on Plusshell 1.0                                         **
;* Copyright 1999/2000 Stephen Clemente, All rights reserved.               **
;*****************************************************************************

        include "LexOS.h"               ;LexOS include file

        xdef    _nostub
        xdef    _ti89
        xdef    _ti92plus

KERNEL_VER      equ     $0100
KERNEL_IDENT    equ     $4C58
VERSION_STRING  macro
                dc.b    "1.0"
                endm

;*****************************************************************************
;* Plusshell header                                                         **
;*****************************************************************************
_begin:
        bra     _main
        cnop    0,4

        dc.b    "68kP"
        dc.w    0
        dc.w    _comment-_begin
        dc.w    _main-_begin
        dc.w    0,3,0,0,0,0,0,0,0,0,0,0

_comment dc.b   "LexOS Kernel 1.0",0
        even

_main:
        movem.l d1-d7/a0-a6,-(a7)       ;Save regs
        bsr     InstallKernel           ;Install kernel
        movem.l (a7)+,d1-d7/a0-a6       ;Restore regs
        rts

InstallKernel:
        tst.l   ($30)                   ;check for kernel
        bne     AlreadyThere            ;kernel already there
        lea.l   ConfigName(pc),a0
        bsr     FindFile
        tst.w   d0
        beq     ConfigNotFound
        _LDEREF d0,a0
        move.w  6(a0),d1
        cmp.w   #$4C43,d1               ;check config header
        bne     ConfigIsBad
        clr.l   d1                      ;clear d1
        move.w  10(a0),d1               ;d1=Number of RAM entries
        move.l  #4,d2                   ;d2=sizeof(long)
        mulu.l  d1,d2                   ;d2=RAM entries*sizeof(long)
        add.l   #_LexOSEnd-_LexOSStart,d2       ;add kernel size to d2
        addq.l  #2,d2                   ;add space for RAM address counter
        move.w  d0,-(a7)                ;save config handle to stack
        move.l  d2,-(a7)                ;push parameter
        move.l  JUMP_TABLE,a0           ;prepare HeapAllocHigh()
        move.l  HeapAllocHigh(a0),a0
        jsr     (a0)                    ;call HeapAllocHigh()
        add.l   #4,a7                   ;restore stack
        move.w  (a7)+,d1                ;restore config handle
        tst.w   d0                      ;test handle
        beq     KernelNoMem             ;not enough memory
        lea.l   KernelHandle(pc),a0     ;get handle backup
        move.w  d0,(a0)                 ;save handle
        _LDEREF d0,a0                   ;a0 points to new buffer
        _LDEREF d1,a1                   ;a1 points to config file
        lea.l   _LexOSStart(pc),a2      ;points to LexOS start
        move.l  #_LexOSEnd-_LexOSStart,d2       ;d2=kernel length
        subq.l  #1,d2                   ;subtract 1 for dbra
KernelCopyLoop:
        move.b  (a2)+,(a0)+             ;copy
        dbra.l  d2,KernelCopyLoop       ;loop
        add.l   #10,a1                  ;add 10 to config pointer
        move.w  (a1)+,d2                ;get number of entries
        move.w  d2,(a0)+                ;copy number of entries
        subq.w  #1,d2                   ;subtract 1 for dbra
RAMCopyLoop:
        move.l  (a1)+,(a0)+             ;copy RAM entry
        dbra.w  d2,RAMCopyLoop          ;loop
        _LDEREF d0,a2                   ;get kernel buffer
        move.w  #KERNEL_VER,($40030)
        move.w  #KERNEL_IDENT,($40032)
        lea.l   exec-_LexOSStart(a2),a0
        move.l  a0,($40034)
        lea.l   Reloc-_LexOSStart(a2),a0
        move.l  a0,($40038)
        lea.l   RelocFromPtr-_LexOSStart(a2),a0
        move.l  a0,($4003c)
        lea.l   Unreloc-_LexOSStart(a2),a0
        move.l  a0,($40040)
        lea.l   UnrelocFromPtr-_LexOSStart(a2),a0
        move.l  a0,($40044)

        lea.l   Success(pc),a0
        move.l  a0,-(a7)
        move.l  JUMP_TABLE,a0
        move.l  ST_helpMsg(a0),a0
        jsr     (a0)
        add.l   #4,a7

        rts

AlreadyThere:
        lea.l   KernelThere(pc),a0
        move.l  a0,-(a7)
        move.l  JUMP_TABLE,a0
        move.l  ST_helpMsg(a0),a0
        jsr     (a0)
        add.l   #4,a7
        rts

ConfigNotFound:
        lea.l   NoConfig(pc),a0
        move.l  a0,-(a7)
        move.l  JUMP_TABLE,a0
        move.l  ST_helpMsg(a0),a0
        jsr     (a0)
        add.l   #4,a7
        rts

ConfigIsBad:
        lea.l   BadConfig(pc),a0
        move.l  a0,-(a7)
        move.l  JUMP_TABLE,a0
        move.l  ST_helpMsg(a0),a0
        jsr     (a0)
        add.l   #4,a7
        rts

KernelNoMem:
        lea.l   BadConfig(pc),a0
        move.l  a0,-(a7)
        move.l  JUMP_TABLE,a0
        move.l  ST_helpMsg(a0),a0
        jsr     (a0)
        add.l   #4,a7
        rts

ConfigName:     dc.b    "ramcfg",0,0,0  ;name of config file

        include "instmsg.inc"           ;Localized installation messages

                even
;*****************************************************************************
;* Main LexOS kernel                                                        **
;*****************************************************************************
_LexOSStart:
;*****************************************************************************
;* LexOS varaible area located at exec()-n                                  **
;*****************************************************************************
KernelID        dc.b    "LX"            ;Kernel ID
KernelVersion   dc.w    $0100           ;Version 1.0
Reserved        dc.l    0               ;Space for a pointer in future version
OrigSP          dc.l    0               ;Old stack-pointer
LastErrCode     dc.l    0
ScreenHandle    dc.w    0               ;Handle of screen backup buffer
_CALCULATOR     dc.w    0               ;Calculator variable
ProgCount       dc.w    0               ;Number of programs currently running
KernelHandle    dc.w    0               ;Handle of kernel in memory

;*****************************************************************************
;* exec() function. It will execute the caller program.                     **
;*****************************************************************************
exec:
        movem.l d1-d7/a0-a6,-(a7)       ;save regs
        move.l  56(a7),a4               ;get return address of program
        sub.l   #4,a4                   ;subtract 4 to get beginning of prog
        move.l  a4,a6                   ;copy address to a6
        bsr     RelocFromPtr            ;call relocate program
        tst.l   d0                      ;test return value
        bne     execError               ;Show error if failed
        lea.l   ProgCount(pc),a0        ;Get number of progs running
        tst.w   (a0)                    ;test
        bne     execNoSaveScreen        ;if not 0 don't save screen
        move.l  #3840,-(a7)             ;allocate screen memory
        move.l  JUMP_TABLE,a0           ;get pointer to HeapAlloc()
        move.l  HeapAlloc(a0),a0
        jsr     (a0)                    ;call it
        add.l   #4,a7                   ;restore stack
        tst.w   d0                      ;test return value
        bne     execCopyScreen          ;if not zero copy screen
        moveq.l #1,d0                   ;else fail with error
        bra     execError
execCopyScreen:
        lea.l   ScreenHandle(pc),a0     ;save screen handle
        move.w  d0,(a0)
        move.l  JUMP_TABLE,a0           ;lock buffer
        move.l  HLock(a0),a0
        move.w  d0,-(a7)                ;push parameter
        jsr     (a0)                    ;call HLock
        add.l   #2,a7                   ;restore stack
        move.l  #LCD_MEM,a1             ;let a1 point to video mem
        move.w  #959,d0
execSaveLoop:
        move.l  (a1)+,(a0)+
        dbra.w  d0,execSaveLoop
        lea.l   ScreenHandle(pc),a0     ;Get screen handle pointer
        move.w  (a0),-(a7)              ;push it onto stack
        move.l  JUMP_TABLE,a0           ;Prepare HeapUnlock()
        move.l  HeapUnlock(a0),a0
        jsr     (a0)                    ;call HeapUnlock()
        add.l   #2,a7                   ;restore stack
        lea.l   OrigSP(pc),a0           ;get pointer to old sp buffer
        move.l  a7,(a0)                 ;save stack pointer
        bsr     InstIntHandlers         ;Install exception handlers
execNoSaveScreen:
        lea.l   ProgCount(pc),a0
        add.w   #1,(a0)                 ;increment program count
        clr.l   d0
        move.w  $c(a6),d0               ;get offset to main
        move.l  56(a7),a6               ;get beginning of program
        sub.l   #4,a6
        add.l   d0,a6                   ;add it to beginning of program
        jsr     (a6)                    ;call program
        move.l  56(a7),a4               ;get beginning of program
        sub.l   #4,a4
        move.l  a4,a6                   ;copy beginning of program to d6
        jsr     UnrelocFromPtr(pc)      ;un-relocate program

        lea.l   ProgCount(pc),a0        ;get pointer to program count
        sub.w   #1,(a0)                 ;decrement it
        bne     execEnd                 ;if not zero goto end
        bsr     RemIntHandlers          ;restore old int vectors
        lea.l   OrigSP(pc),a0           ;get pointer to old sp buffer
        move.l  #0,(a0)                 ;set to zero
        lea.l   LastErrCode(pc),a0      ;get ptr to last error code
        clr.l   (a0)                    ;set to zero
        lea.l   ScreenHandle(pc),a0     ;get pointer to screen handle
        move.w  (a0),-(a7)              ;push handle onto stack as parameter
        move.l  JUMP_TABLE,a0           ;prepare HeapDeref()
        move.l  HeapDeref(a0),a0
        jsr     (a0)                    ;call HeapDeref()
        add.l   #2,a7                   ;restore stack
        move.l  #LCD_MEM,a1             ;let a1 point to video mem
        move.w  #959,d0
execRestoreLoop:
        move.l  (a0)+,(a1)+
        dbra.w  d0,execRestoreLoop

        lea.l   ScreenHandle(pc),a0     ;get pointer to screen handle
        move.w  (a0),-(a7)              ;push handle onto stack as parameter
        move.l  JUMP_TABLE,a0           ;prepare HeapFree()
        move.l  HeapFree(a0),a0
        jsr     (a0)                    ;call HeapFree()
        add.l   #2,a7                   ;restore stack

        move.l  d0,-(a7)                ;save d0
        bsr     CleanUpSys              ;unrelocate relocated libs
        move.l  (a7)+,d0                ;restore d0

execEnd:
        movem.l (a7)+,d1-d7/a0-a6       ;restore regs
        add.l   #4,a7
        rts                             ;return

;*****************************************************************************
;* Note: This code is called in case of a system crash.                     **
;*****************************************************************************
AbnormalTermination:
        lea.l   LastErrCode(pc),a0      ;get ptr to last error code
        move.l  d0,(a0)                 ;save error code
        bsr     RemIntHandlers          ;restore old int vectors
        lea.l   OrigSP(pc),a0           ;get pointer to old sp buffer
        move.l  #0,(a0)                 ;set to zero
        bsr     ShowExceptionMessage    ;display error message
        lea.l   ProgCount(pc),a0        ;get program count
        clr.w   (a0)                    ;set program count to zero
        lea.l   ScreenHandle(pc),a0     ;get pointer to screen handle
        move.w  (a0),-(a7)              ;push handle onto stack as parameter
        move.l  JUMP_TABLE,a0           ;prepare HeapDeref()
        move.l  HeapDeref(a0),a0
        jsr     (a0)                    ;call HeapDeref()
        add.l   #2,a7                   ;restore stack
        move.l  #LCD_MEM,a1             ;let a1 point to video mem
        move.w  #959,d0
ATRestoreLoop:
        move.l  (a0)+,(a1)+
        dbra.w  d0,ATRestoreLoop

        lea.l   ScreenHandle(pc),a0     ;get pointer to screen handle
        move.w  (a0),-(a7)              ;push handle onto stack as parameter
        move.l  JUMP_TABLE,a0           ;prepare HeapFree()
        move.l  HeapFree(a0),a0
        jsr     (a0)                    ;call HeapFree()
        add.l   #2,a7                   ;restore stack

        move.l  d0,-(a7)                ;save d0
        bsr     CleanUpSys              ;unrelocate unrelocated libs
        move.l  (a7)+,d0                ;restore d0

        movem.l (a7)+,d1-d7/a0-a6       ;restore regs
        add.l   #4,a7
        rts                             ;return


;*****************************************************************************
;* Display error messages here                                              **
;* d0 can contain the following error codes:                                **
;* 1: Not enough memory * 2: Library not found * 3: Wrong library version   **
;*****************************************************************************
execError:
        cmp.l   #1,d0                   ;Test if not enough mem
        beq     execNoMem
        cmp.l   #2,d0                   ;Test if lib missing
        beq     execNoLib
        cmp.l   #3,d0                   ;Test if wrong lib version
        beq     execLibVer

;** Display 'Bad File Format' error message **********************************
execBadFormat:
        lea.l   BadFileFormatMsg(pc),a0
        move.l  a0,-(a7)
        move.l  JUMP_TABLE,a0
        move.l  ST_helpMsg(a0),a0
        jsr     (a0)
        add.l   #4,a7
        bra     execEnd

;** Display 'Not enough memory' error message ********************************
execNoMem:
        lea.l   NoMemMsg(pc),a0
        move.l  a0,-(a7)
        move.l  JUMP_TABLE,a0
        move.l  ST_helpMsg(a0),a0
        jsr     (a0)
        add.l   #4,a7
        bra     execEnd

;** Display 'Library not found' error message ********************************
execNoLib:
        lea.l   MissingLibName(pc),a1
        move.l  (a0)+,(a1)+
        move.l  (a0)+,(a1)+
        lea.l   MissingLibMsg(pc),a0
        move.l  a0,-(a7)
        move.l  JUMP_TABLE,a0
        move.l  ST_helpMsg(a0),a0
        jsr     (a0)
        add.l   #4,a7
        bra     execEnd

;** Display 'Wrong library version' error message ****************************
execLibVer:
        lea.l   WrongLibVersionName(pc),a1
        move.l  (a0)+,(a1)+
        move.l  (a0)+,(a1)+
        lea.l   WrongLibVersionMsg(pc),a0
        move.l  a0,-(a7)
        move.l  JUMP_TABLE,a0
        move.l  ST_helpMsg(a0),a0
        jsr     (a0)
        add.l   #4,a7
        bra     execEnd

;*****************************************************************************
;* Exeption handlers                                                        **
;*****************************************************************************
EXBusError      equ     $00
EXAddrError     equ     $01
EXIllegalError  equ     $02
EXDivideError   equ     $03
EXLine1010Error equ     $04
EXLine1111Error equ     $05
EXPrivError     equ     $06
EXCHKError      equ     $07

;*****************************************************************************
;* Install exception handlers                                               **
;*****************************************************************************
InstIntHandlers:
        lea.l   OldIntVec(pc),a0
        move.l  $40008,(a0)+            ;save bus error vector
        move.l  $4000C,(a0)+            ;save address error vector
        move.l  $40010,(a0)+            ;save illegal error vector
        move.l  $40014,(a0)+            ;save zero div error vector
        move.l  $40028,(a0)+            ;save line 1010 error vector
        move.l  $4002C,(a0)+            ;save line 1111 error vector
        move.l  $40020,(a0)+            ;save privilige violation vector
        move.l  #$40064,a1              ;let a1 point to 1st auto int
        move.l  (a1)+,(a0)+             ;save all 7 auto ints
        move.l  (a1)+,(a0)+
        move.l  (a1)+,(a0)+
        move.l  (a1)+,(a0)+
        move.l  (a1)+,(a0)+
        move.l  (a1)+,(a0)+
        move.l  (a1)+,(a0)+
        lea.l   doBusError(pc),a0
        move.l  a0,$40008
        lea.l   doAddrError(pc),a0
        move.l  a0,$4000C
        lea.l   doIllegalError(pc),a0
        move.l  a0,$40010
        lea.l   doDivideError(pc),a0
        move.l  a0,$40014
        lea.l   doLine1010Error(pc),a0
        move.l  a0,$40028
        lea.l   doLine1111Error(pc),a0
        move.l  a0,$4002C
        lea.l   doPrivError(pc),a0
        move.l  a0,$40020
        rts

RemIntHandlers:
        lea.l   OldIntVec(pc),a0
        move.l  (a0)+,$40008            ;restore bus error vector
        move.l  (a0)+,$4000C            ;restore address error vector
        move.l  (a0)+,$40010            ;restore illegal error vector
        move.l  (a0)+,$40014            ;restore zero div error vector
        move.l  (a0)+,$40028            ;restore line 1010 error vector
        move.l  (a0)+,$4002C            ;restore line 1111 error vector
        move.l  (a0)+,$40020            ;restore privilige violation vector
        move.l  #$40064,a1              ;let a1 point to 1st auto int
        move.l  (a0)+,(a1)+             ;restore all 7 auto ints
        move.l  (a0)+,(a1)+
        move.l  (a0)+,(a1)+
        move.l  (a0)+,(a1)+
        move.l  (a0)+,(a1)+
        move.l  (a0)+,(a1)+
        move.l  (a0)+,(a1)+
        rts

doBusError:
        move.l  d0,-(a7)
        move.l  #EXBusError,d0
        bra     ExceptionHandle
doAddrError:
        move.l  d0,-(a7)
        move.l  #EXAddrError,d0
        bra     ExceptionHandle
doIllegalError:
        move.l  d0,-(a7)
        move.l  #EXIllegalError,d0
        bra     ExceptionHandle
doDivideError:
        move.l  d0,-(a7)
        move.l  #EXDivideError,d0
        bra     ExceptionHandle
doLine1010Error:
        move.l  d0,-(a7)
        move.l  #EXLine1010Error,d0
        bra     ExceptionHandle
doLine1111Error:
        move.l  d0,-(a7)
        move.l  #EXLine1111Error,d0
        bra     ExceptionHandle
doPrivError:
        move.l  d0,-(a7)
        move.l  #EXPrivError,d0
        bra     ExceptionHandle


;* Process exception *********************************************************
ExceptionHandle:
        move.l  a0,-(a7)                ;Save a0
        lea.l   RegisterDump(pc),a0     ;Let a0 point to register buffer
        move.l  (a7)+,32(a0)            ;save a0
        move.l  (a7)+,(a0)              ;save d0
        movem.l d1-d7,4(a0)             ;save d1-d7
        movem.l a1-a6,36(a0)            ;save a1-a6
        move.l  2(a7),64(a0)            ;save pc
        move.w  (a7),68(a0)             ;save sr
        lea.l   ProcessException(pc),a0
        move.l  a0,2(a7)                ;modify return address
        rte                             ;return from interrupt handler
ProcessException:
        lea.l   RegisterDump(pc),a0
        move.l  a7,60(a0)               ;save stack pointer
        lea.l   OrigSP(pc),a0           ;restore stack-pointer
        move.l  (a0),a7
        bra     AbnormalTermination     ;end program execution

;*****************************************************************************
;* Display exception message                                                **
;*****************************************************************************
ShowExceptionMessage:
        lea.l   LastErrCode(pc),a0
        move.l  (a0),d0
        tst.l   d0
        beq     showBusError
        sub.l   #1,d0
        beq     showAddrError
        sub.l   #1,d0
        beq     showIllegalError
        sub.l   #1,d0
        beq     showDivideError
        sub.l   #1,d0
        beq     showLine1010Error
        sub.l   #1,d0
        beq     showLine1111Error
        bra     showPrivError

showBusError:
        lea.l   msgBusError(pc),a0
        bra     ShowErrMessage
showAddrError:
        lea.l   msgAddrError(pc),a0
        bra     ShowErrMessage
showIllegalError:
        lea.l   msgIllegalError(pc),a0
        bra     ShowErrMessage
showDivideError:
        lea.l   msgDivideError(pc),a0
        bra     ShowErrMessage
showLine1010Error:
        lea.l   msgLine1010Error(pc),a0
        bra     ShowErrMessage
showLine1111Error:
        lea.l   msgLine1111Error(pc),a0
        bra     ShowErrMessage
showPrivError:
        lea.l   msgPrivError(pc),a0
        bra     ShowErrMessage


ShowErrMessage:

ShowBluescreen:
        move.l  a0,-(a7)                        ;save ptr to error message
        move.w  #LCD_DISPLAY/8,IO_VIDEOMEM      ;set video ram to standard value
        move.l  #LCD_DISPLAY,a0
        move.w  #959,d0                         ;init number of loops
ShowBluescreenLoop:
        clr.l   (a0)+                           ;clear screen
        dbra.w  d0,ShowBluescreenLoop

        move.w  #0,-(a7)                ;set font to #0
        move.l  JUMP_TABLE,a0           ;prepare FontSetSys()
        move.l  FontSetSys(a0),a0
        jsr     (a0)                    ;call FontSetSys()
        add.l   #2,a7                   ;restore stack

        move.l  (a7)+,a0                ;get pointer to error message
        move.w  #1,-(a7)                ;write mode
        move.l  a0,-(a7)                ;string
        move.w  #8,-(a7)                ;y
        move.w  #0,-(a7)                ;x
        move.l  JUMP_TABLE,a0
        move.l  DrawStr(a0),a0
        jsr     (a0)
        add.l   #10,a7                  ;restore stack

        lea.l   RegisterDump(pc),a0     ;get pointer to register dump
        move.l  64(a0),d0               ;put old PC into d0
        lea.l   errLine1(pc),a0         ;get pointer to error line
        add.l   #14,a0                  ;add offset to hex number
        bsr     MakeHEX24               ;generate hex number
        sub.l   #14,a0                  ;restore pointer
        move.w  #1,-(a7)                ;write mode
        move.l  a0,-(a7)                ;string
        move.w  #0,-(a7)                ;y
        move.w  #0,-(a7)                ;x

        lea.l   RegisterDump(pc),a0
        move.l  64(a0),a0
        bsr     GetModuleAddress
        cmp.l   #0,a0
        beq     UnknownModName
        move.l  a0,-(a7)
        lea.l   RegisterDump(pc),a1
        move.l  64(a1),d0
        move.w  12(a0),d1
        _LDEREF d1,a0
        sub.l   a0,d0
        lea.l   errLine1(pc),a0
        add.l   #22,a0
        bsr     MakeHEX24
        move.l  (a7)+,a0
        lea.l   modName(pc),a1
        move.l  (a0)+,(a1)+
        move.l  (a0),(a1)
        bra     ModNameDone
UnknownModName:
        lea.l   defModName(pc),a0
        lea.l   modName(pc),a1
        move.l  (a0)+,(a1)+
        move.l  (a0),(a1)

        lea.l   RegisterDump(pc),a1
        move.l  64(a1),d0
        lea.l   errLine1(pc),a0
        add.l   #22,a0
        bsr     MakeHEX24
ModNameDone:
        move.l  JUMP_TABLE,a0
        move.l  DrawStr(a0),a0
        jsr     (a0)
        add.l   #10,a7
        move.l  #0,d0
        bsr     PrintRegs
        move.l  #1,d0
        bsr     PrintRegs
        move.l  #2,d0
        bsr     PrintRegs
        move.l  #3,d0
        bsr     PrintRegs
        move.l  #4,d0
        bsr     PrintRegs
        move.l  #5,d0
        bsr     PrintRegs
        move.l  #6,d0
        bsr     PrintRegs
        move.l  #7,d0
        bsr     PrintRegs
;Print help lines
        lea.l   errLine3(pc),a0
        move.w  #1,-(a7)                ;write mode
        move.l  a0,-(a7)                ;string
        move.w  #80,-(a7)               ;y
        clr.w   -(a7)                   ;x
        move.l  JUMP_TABLE,a0
        move.l  DrawStr(a0),a0
        jsr     (a0)
        add.l   #10,a7
        lea.l   errLine4(pc),a0
        move.w  #1,-(a7)                ;write mode
        move.l  a0,-(a7)                ;string
        move.w  #88,-(a7)               ;y
        clr.w   -(a7)                   ;x
        move.l  JUMP_TABLE,a0
        move.l  DrawStr(a0),a0
        jsr     (a0)
        add.l   #10,a7

        move.l  JUMP_TABLE,a0
        move.l  ngetchx(a0),a0
        jsr     (a0)
        rts

;*****************************************************************************
;* Creates a 32 bit hex string from d0 at (a0).                             **
;*****************************************************************************
MakeHEX:
        movem.l d0-d2/a1,-(a7)          ;save regs
        move.w  #7,d2                   ;loop counter
        lea.l   HexLUT(pc),a1           ;a1 points to HexLUT
        add.l   #8,a0                   ;add 8 to string pointer
MakeHEXLoop1:
        move.w  d0,d1                   ;copy value to d1
        and.w   #$000F,d1               ;isolate lower nibble
        move.b  0(a1,d1.w),-(a0)        ;copy character
        lsr.l   #4,d0                   ;shift d0
        dbra.w  d2,MakeHEXLoop1         ;loop
        movem.l (a7)+,d0-d2/a1          ;restore regs
        rts                             ;rts

;*****************************************************************************
;* Creates a 24 bit hex string from d0 at (a0).                                    **
;*****************************************************************************
MakeHEX24:
        movem.l d0-d2/a1,-(a7)          ;save regs
        move.w  #5,d2                   ;loop counter
        lea.l   HexLUT(pc),a1           ;a1 points to HexLUT
        add.l   #6,a0                   ;add 8 to string pointer
MakeHEX24Loop1:
        move.w  d0,d1                   ;copy value to d1
        and.w   #$000F,d1               ;isolate lower nibble
        move.b  0(a1,d1.w),-(a0)        ;copy character
        lsr.l   #4,d0                   ;shift d0
        dbra.w  d2,MakeHEXLoop1         ;loop
        movem.l (a7)+,d0-d2/a1          ;restore regs
        rts                             ;rts


PrintRegs:                              ;d0=reg number
        move.w  d0,-(a7)                ;save d0
        lea.l   errLine2(pc),a0         ;get pointer to error line
        add.b   #48,d0                  ;generate ASCII code of line
        move.b  d0,1(a0)                ;fill-in register numbers
        move.b  d0,17(a0)
        lea.l   RegisterDump(pc),a2     ;register dump
        move.w  (a7),d0                 ;get register number from stack
        lsl.w   #2,d0                   ;multiply by 4
        and.l   #$FFFF,d0               ;isolate lower word
        add.l   d0,a2                   ;address of DReg
        move.l  (a2),d0
        add.l   #3,a0
        bsr     MakeHEX
        move.l  32(a2),d0
        add.l   #16,a0
        bsr     MakeHEX
        lea.l   errLine2(pc),a0
        move.w  (a7)+,d0
        move.w  #1,-(a7)                ;write mode
        move.l  a0,-(a7)                ;string
        lsl.l   #3,d0
        add.l   #16,d0
        move.w  d0,-(a7)                ;y
        clr.w   -(a7)                   ;x
        move.l  JUMP_TABLE,a0
        move.l  DrawStr(a0),a0
        jsr     (a0)
        add.l   #10,a7
        rts
        
;*****************************************************************************
;* SYMENTRY *GetModuleAddress(void *PC=a0);                                 **
;* Returns a pointer to the file which occupies the space at a0. If there is**
;* none NULL will be returned.                                              **
;*****************************************************************************
GetModuleAddress:
        move.l  a0,-(a7)                ;save pointer
        move.w  #$0A,-(a7)              ;magic number for SymFindFirst()
        move.l  #0,-(a7)                ;magic number for SymFindFirst()
        move.l  JUMP_TABLE,a0           ;Prepare SymFindFirst()
        move.l  SymFindFirst(a0),a0
        jsr     (a0)                    ;call SymFindFirst()
        add.l   #6,a7                   ;restore stack
GetModuleAddressLoop:
        move.l  JUMP_TABLE,a0           ;Prepare SymFindNext()
        move.l  SymFindNext(a0),a0
        jsr     (a0)                    ;call SymFindNext()
        cmp.l   #0,a0                   ;test return value
        beq     GetModuleAddressNone    ;no more files
        move.l  8(a0),d0                ;get attributes
        and.l   #$00000280,d0           ;isolate folder and archive bit
        tst.l   d0                      ;if set
        bne     GetModuleAddressLoop    ;then process next
        move.w  12(a0),d0               ;get file handle
        _LDEREF d0,a1
        move.w  (a1),d0                 ;get length
        and.l   #$FFFF,d0               ;isolate lower word
        move.l  (a7),d1                 ;get pointer
        cmp.l   a1,d1                   ;compare
        blt     GetModuleAddressLoop    ;address too low
        add.l   d0,a1                   ;add length of file to pointer
        cmp.l   a1,d1                   ;compare
        bgt     GetModuleAddressLoop    ;address too high
        add.l   #4,a7                   ;restore stack
        rts                             ;return
GetModuleAddressNone:
        add.l   #4,a7
        move.l  #0,a0
        rts


;*****************************************************************************
;* Relocation code                                                          **
;*****************************************************************************
Reloc:
        movem.l d1-d7/a1-a6,-(a7)       ;save regs
        _LDEREF d0,a6                   ;dereference handle
        bra     RelocFromPtrStart

RelocFromPtr:
        movem.l d1-d7/a1-a6,-(a7)       ;save regs
RelocFromPtrStart:
        move.l  4(a6),d0                ;get file ID
        cmp.l   #$36386b41,d1           ;compare with 68kA
        beq     RelocDoneWorking        ;don't relocate
        and.l   #$ffffff00,d0           ;get the first 3 characters
        cmp.l   #$36386b00,d0           ;compare with '68k'
        bne     RelocBadFileFormat      ;fail if wrong ID
        move.l  a6,a4                   ;let a4 point to program
        tst.w   $8(a6)                  ;check reloc count
        bne     RelocDoneWorking        ;if not zero don't relocate
        or.w    #$8000,$8(a6)           ;set top bit of reloc count
        clr.w   $12(a6)                 ;clear BSS handle
        clr.l   d0                      ;clear d0
        move.w  $14(a4),d0              ;let d0 contain offset to BSS table
        tst.w   d0                      ;test offset
        beq     RelocNoBSSAlloc         ;if 0 then skip BSS allocation
        add.l   d0,a4                   ;get length of BSS table
        tst.l   (a4)                    ;if zero
        beq     RelocNoBSSAlloc         ;then skip BSS allocation
        move.l  (a4),d0                 ;put length of BSS into d0
        movem.l a4-a6,-(a7)             ;save regs
        move.l  d0,-(a7)                ;push length onto stack
        move.l  JUMP_TABLE,a4           ;prepare HeapAlloc()
        move.l  HeapAlloc(a4),a4
        jsr     (a4)                    ;call HeapAlloc()
        add.l   #4,a7                   ;restore stack
        movem.l (a7)+,a4-a6             ;restore regs
        tst.w   d0                      ;test result
        beq     RelocNoMem              ;fail if not enough memory
        move.w  d0,$12(a6)              ;store BSS handle
        _LOCK   d0                      ;lock the BSS block so it won't move
RelocNoBSSAlloc:
        move.l  a6,a4                   ;let a4 point to beginning of program
        add.l   #$1a,a4                 ;let a4 point to import table
        move.w  (a4)+,d7                ;put number of libs into d7
RelocLibCheckLoop:
        tst.w   d7                      ;if remaining libs=0
        beq     RelocLibCheckEnd        ;finished checking libs
        move.l  a4,a0                   ;let a0 point to lib name
        bsr     GetFileAddress          ;get the file
        tst.w   d0                      ;if handle=0
        beq     RelocLibNotFound        ;then fail
        move.l  6(a0),d0                ;let d0 contain file ID
        and.l   #$ffffff00,d0           ;get first 3 chars
        cmp.l   #$36386b00,d0           ;compare with '68k'
        bne     RelocLibNotFound        ;if not equal then fail
        add.l   #10,a4                  ;let a0 point to next lib
        dbra    d7,RelocLibCheckLoop    ;loop
RelocLibCheckEnd:
        move.l  a6,a4                   ;let a4 point to beginning of program
        add.l   #$1a,a4                 ;let a4 point to import table
        move.w  (a4)+,d7                ;put number of libs into d7
        move.w  d7,d6                   ;copy d7 to d6
        mulu.w  #10,d6                  ;multiply number of libs by 10
        move.l  a4,a3                   ;let a3 point to lib name
        add.l   d6,a3                   ;let a3 point to lib reloc table
RelocLibLoop:
        tst.w   d7                      ;if remaining libs=0
        beq     RelocLibEnd             ;done
        move.l  a4,a0                   ;let a0 point to libname
        bsr     GetFileAddress          ;get the file
        tst.w   d0                      ;if handle=0
        beq     RelocLibNotFound        ;fail
        _LOCK   d0                      ;lock lib
        move.w  d0,d1                   ;copy handle to d1
        movem.l d1-d7/a0-a6,-(a7)       ;save regs
        move.l  a0,a6                   ;let a6 point to lib file
        add.l   #2,a6                   ;add 2 to let a6 point to lib itself
        bsr     RelocFromPtr            ;relocate lib (recursively)
        movem.l (a7)+,d1-d7/a0-a6       ;restore regs
        tst.w   d0                      ;if result!=0
        bne     RelocLibNotFound        ;fail
        _LDEREF d1,a2                   ;let a2 point to lib file
        add.l   #2,a2                   ;add 2 to let a6 point to lib itself
        move.l  a2,a5                   ;let a5 point to lib aswell
        clr.l   d0                      ;clear d0
        move.w  $16(a2),d0              ;d0=offset to lib's export-table
        add.l   d0,a2                   ;let a2 point to lib's export table
        move.w  #1,d3                   ;set relocation-type (d3) to 1
        bsr     DoReloc                 ;do the actual relocation
        tst.w   d0                      ;test result
        bne     RelocWrongLibVersion    ;fail because of wrong lib version
        add.l   #10,a4                  ;let a4 point to next lib entry
        dbra    d7,RelocLibLoop         ;loop
        bra     RelocLibEnd             ;skip following code
RelocLibErr:

RelocLibEnd:
        tst.w   (a3)+                   ;check for ROM calls
        beq     RelocNoROM              ;skip ROM relocation if there are none
        move.l  $c8,a2                  ;let a2 point to ROM call table
        sub.w   #2,a2                   ;subtract 2
        move.w  #2,d3                   ;set relocation-type (d3) to 2
        bsr     DoReloc                 ;do the actual relocation
RelocNoROM:
        tst.w   (a3)+                   ;test RAM relocation table
        beq     RelocNoRAM              ;skip RAM relocation if there are none
        clr.l   d0                      ;clear d0
        move.w  $18(a6),d0              ;d0=offset to extra RAM table
        move.l  a6,a4                   ;let a4 point to program
        add.l   d0,a4                   ;let a4 point to extra RAM table
        move.l  ($4),d0                 ;load interupt vector $01 into d0
        and.l   #$400000,d0             ;check to see whether it's a TI-89
        beq     Reloc89                 ;if TI-89 then relocate for a TI-89
        lea.l   _CALCULATOR(pc),a0      ;let a0 point to calculator type var.
        move.b  #1,(a0)                 ;set type to 1(=TI-92+)
        lea.l   RAMTable(pc),a2       ;let a0 point to TI-92+ RAM table
        lea.l   2(a4),a4                ;let a4 point to RAM table+2
        bra     RelocDoRAM              ;skip TI-89 part
Reloc89:
        lea.l   _CALCULATOR(pc),a0      ;let a0 point to calculator type var.
        move.b  #0,(a0)                 ;set type to 0(=TI-89)
        lea.l   RAMTable(pc),a2       ;let a2 point to ram table
RelocDoRAM:
        move.w  #2,d3                   ;set relocation-type (d3) to 2
        bsr     DoReloc                 ;actual relocation
RelocNoRAM:
        clr.l   d0                      ;clear d0
        move.w  (a3)+,d0                ;get first offset of reloc table
        tst.w   d0                      ;check entry
        beq     RelocRelEnd             ;if zero then don't relocate
        move.l  a6,a0                   ;let a0 point to program
        add.l   d0,a0                   ;add offset of reloc entry
        move.l  a6,d0                   ;let d0 contain pointer to program
        add.l   d0,(a0)                 ;add program offset to program pointer
        bra     RelocNoRAM              ;loop
RelocRelEnd:
        tst.w   $14(a6)                 ;check for BSS table
        beq     RelocDone               ;skip BSS relocation if zero
        tst.w   $12(a6)                 ;check for BSS buffer
        beq     RelocDone               ;skip BSS relocation if zero
        add.l   #4,a3                   ;let a3 point to BSS entries
        move.w  $12(a6),d0              ;load BSS handle into d0
        _LDEREF d0,a1                   ;let a1 point to BSS buffer
RelocBSSLoop:
        clr.l   d0                      ;clear d0
        move.w  (a3)+,d0                ;load offset into d0
        tst.w   d0                      ;test d0
        beq     RelocDone               ;if zero then finished
        move.l  a6,a0                   ;let a0 point to beginning of program
        add.l   d0,a0                   ;add offset to d0
        move.l  a1,d0                   ;let d0 contain pointer to BSS block
        add.l   d0,(a0)                 ;add BSS address to ptr in program
        bra     RelocBSSLoop            ;loop
RelocDone:
        move.w  $8(a6),d0               ;get reloc count
        and.w   #$7FFF,d0               ;get rid of working bit
        add.w   #1,d0                   ;add 1 to reloc count
        and.w   #$8000,$8(a6)           ;zero out reloc count
        or.w    d0,$8(a6)               ;copy reloc count back to program
RelocOK:
        clr.l   d0                      ;let d0=0 (0=success)
RelocEnd:
        and.w   #$7FFF,$8(a6)           ;clear working bit
        movem.l (a7)+,d1-d7/a1-a6       ;restore regs
        rts                             ;return
RelocDoneWorking:
        move.w  $8(a6),d0               ;get reloc count
        and.w   #$7FFF,d0               ;get rid of working bit
        add.w   #1,d0                   ;add 1 to reloc count
        and.w   #$8000,$8(a6)           ;zero out reloc count
        or.w    d0,$8(a6)               ;copy reloc count back to program
        clr.l   d0                      ;let d0=0 (0=success)
        movem.l (a7)+,d1-d7/a1-a6       ;restore regs
        rts                             ;return

RelocBadFileFormat:
        moveq.l #4,d0           ;Put fail value into d0
        bra     RelocEnd        ;go to end

RelocLibNotFound:
        bsr     DeallocBSS      ;dealloc BSS
        move.l  a4,a0           ;put lib name into a0
        moveq.l #2,d0           ;put fail value into d0
        bra     RelocEnd        ;go to end

RelocNoMem:
        moveq.l #1,d0           ;put fail value into d0
        bra     RelocEnd        ;go to end

RelocWrongLibVersion:
        bsr     DeallocBSS      ;dealloc BSS
        move.l  a4,a0           ;put lib name into a0
        moveq.l #3,d0           ;put fail value into d0
        bra     RelocEnd        ;go to end

Unreloc:
        _LDEREF  d0,a6           ;dereference handle
        movem.l d1-d7/a1-a6,-(a7)
        bra     UnrelocFromPtrStart

UnrelocFromPtr:
        movem.l d1-d7/a1-a6,-(a7)
UnrelocFromPtrStart:
        tst.w   $8(a6)          ;if reloc count=0
        beq     UnrelocOK       ;then finish
        sub.w   #1,$8(a6)       ;subtract 1 from reloc count
        move.l  a6,a4           ;let a4 point to program
        tst.w   $8(a6)          ;test reloc count
        bne     UnrelocOK       ;if not zero then finish
        add.l   #$1a,a4         ;let a4 point to import table
        move.w  (a4)+,d7        ;let d7 contian number of libs
        move.w  d7,d6           ;copy number of libs to d6
        mulu.w  #10,d6          ;multiply number of libs by 10
        move.l  a4,a3           ;let a3 point to import table
        add.l   d6,a3           ;let a3 point to lib reloc table
UnrelocLibLoop:
        tst.w   d7              ;if d7=0
        beq     UnrelocLibEnd   ;then finished
        move.l  a4,a0           ;let a0 point to file name
        bsr     GetFileAddress  ;call GetFileAddress()
        tst.w   d0              ;test return value
        beq     UnrelocEnd      ;if zero then don't unreloc
        move.w  d0,d1           ;copy handle to d1
        movem.l d1-d7/a0-a6,-(a7)       ;save regs
        move.l  a0,a6           ;let a6 point to library file
        add.l   #2,a6           ;let a6 point to lib itself
        bsr     UnrelocFromPtr  ;actual unrelocation
        movem.l (a7)+,d1-d7/a0-a6       ;restore regs
        tst.w   d0              ;check return value
        bne     UnrelocLibEnd   ;skip following lines
        _LDEREF d1,a2           ;let a2 point to library file
        add.l   #2,a2           ;let a2 point to actual library
        tst.w   $8(a2)          ;test reloc count
        bne     DontUnlockLib   ;if not zero don't unlock lib
        _UNLOCK d1
DontUnlockLib:
        move.l  a2,a5           ;copy pointer to a5
        clr.l   d0              ;clear d0
        move.w  $16(a2),d0      ;let d0 contain offset to lib's export table
        add.l   d0,a2           ;let a2 point to export table
        move.w  #1,d3           ;set relocation mode to 1
        bsr     DoUnreloc       ;actual derelocation
        add.l   #10,a4          ;let a4 point to next library
        dbra    d7,UnrelocLibLoop       ;loop
UnrelocLibEnd:
        tst.w   (a3)+           ;test if ROM calls are made
        beq     UnrelocNoROM    ;skip next if zero
        move.l  $c8,a2          ;let a2 point to ROM call table
        sub.w   #2,a2           ;subtract 2
        move.w  #2,d3           ;set relocation mode to 2
        bsr     DoUnreloc       ;actual derelocation
UnrelocNoROM:
        tst.w   (a3)+           ;check for ram table
        beq     UnrelocNoRAM    ;skip next code if none
        clr.l   d0
        move.w  $18(a6),d0
        move.l  a6,a4
        add.l   d0,a4
        move.l  ($4),d0
        and.l   #$400000,d0
        beq     Unreloc89
        lea.l   _CALCULATOR(pc),a0
        move.b  #1,(a0)                         ;1=CALC_TI92PLUS
        lea.l   RAMTable(pc),a2
        lea.l   2(a4),a4
        bra     UnrelocDoRAM
Unreloc89:
        lea.l   _CALCULATOR(pc),a0
        move.b  #0,(a0)                         ;0=CALC_TI89
        lea.l   RAMTable(pc),a2
UnrelocDoRAM:
        move.w  #2,d3
        bsr     DoUnreloc
UnrelocNoRAM:
        clr.l   d0
        move.w  (a3)+,d0
        tst.w   d0
        beq     UnrelocRelEnd
        move.l  a6,a0
        add.l   d0,a0
        move.l  a6,d0
        sub.l   d0,(a0)
        bra     UnrelocNoRAM
UnrelocRelEnd:
        tst.w   $14(a6)
        beq     UnrelocOK
        add.l   #4,a3
        move.w  $12(a6),d0
        tst.w   d0
        beq     UnrelocOK
        _LDEREF d0,a1
UnrelocBSSLoop:
        clr.l   d0
        move.w  (a3)+,d0
        tst.w   d0
        beq     UnrelocBSSDone
        move.l  a6,a0
        add.l   d0,a0
        move.l  a1,d0
        sub.l   d0,(a0)
        bra     UnrelocBSSLoop
UnrelocBSSDone:
        bsr     DeallocBSS
UnrelocOK:
        clr.l   d0
UnrelocEnd:
        movem.l (a7)+,d1-d7/a1-a6
        rts

DeallocBSS:
        tst.w   $12(a6)
        beq     NoBSS
        move.l  a4,-(a7)
        move.w  $12(a6),-(a7)           ;push BSS handle onto stack
        move.l  JUMP_TABLE,a4
        move.l  HeapFree(a4),a4
        jsr     (a4)
        add.l   #2,a7
        clr.w   $12(a6)
        move.l  (a7)+,a4
NoBSS:
        rts

;*****************************************************************************
;* DoReloc()                                                                **
;* a3 points to reloc table, a2 points to export table                      **
;* a4=extra ram table?                                                      **
;* a6 points to destination, a5 to source. d3 is the relocation mode.       **
;* 1) RelocLib, 2) RelocROM/RAM                                             **
;*****************************************************************************
DoReloc:
        clr.l   d0                      ;clear d0
DoRelocStart:
        move.w  (a2)+,d5                ;d5=number of functions in export table
        move.w  (a3)+,d6                ;d6=number of functions in import table
DoRelocLoop:        
        clr.l   d4                      ;clear d0
        move.w  (a3)+,d4                ;d4=number of first function
        move.l  a2,a1                   ;let a1 point to export data
        clr.l   d2                      ;clear d2
        cmp.w   #$8000,d4               ;if function number<=0x8000
        bcs     DoRelocNotWord          ;skip following two instructions
        moveq.l #1,d2                   ;let d2=1
        and.w   #$7fff,d4               ;remove highest bit from d4
DoRelocNotWord:
        cmp.w   #$4000,d4               ;if number<=0x4000
        bcs     DoRelocNotExtra         ;skip the following section
        move.l  a4,a1                   ;let a1 point to extra ram table
        and.w   #$3fff,d4               ;remove highest 2 bits from d4
        lsl.l   #2,d4                   ;multiply d4 by 4
        add.l   d4,a1                   ;add d4 to export data pointer
        move.w  (a1),a1                 ;move the word at a1 to a1
        bra     DoRelocFuncLoop         ;loop
DoRelocNotExtra:
        cmp.w   d5,d4                   ;test if function is is export table
        bgt     DoRelocError            ;if not then quit with error
        lsl.l   d3,d4                   ;shift d4 by relocation mode
        add.l   d4,a1                   ;add result to export data pointer
        cmp.w   #2,d3                   ;if reloc mode=2
        beq     DoRelocLoopDWordOfs     ;skip following code
        move.w  (a1),a1
        add.l   a5,a1
        bra     DoRelocFuncLoop
DoRelocLoopDWordOfs:
        move.l  (a1),a1
DoRelocFuncLoop:
        clr.l   d4
        move.w  (a3)+,d4
        tst.w   d4
        beq     DoRelocFuncEnd
        move.l  a6,a0
        add.l   d4,a0
        move.l  a1,d4
        tst.l   d2
        bne     DoRelocAddWord
        tst.l   d0
        bne     DoRelocSubDWord
        add.l   d4,(a0)
        bra     DoRelocFuncLoop
DoRelocSubDWord:
        sub.l   d4,(a0)
        bra     DoRelocFuncLoop
DoRelocAddWord:
        tst.l   d0
        bne     DoRelocSubWord
        add.w   d4,(a0)
        bra     DoRelocFuncLoop
DoRelocSubWord:
        sub.w   d4,(a0)
        bra     DoRelocFuncLoop
DoRelocFuncEnd:
        dbra    d6,DoRelocLoop
        clr.l   d0
        rts
DoRelocError:
        moveq.l #1,d0
        rts

DoUnreloc:
        moveq.l #1,d0
        bra     DoRelocStart


;*****************************************************************************
;* Unrelocate all modules that should not be relocated.                     **
;*****************************************************************************
CleanUpSys:
        move.w  #$0A,-(a7)              ;magic number for SymFindFirst()
        move.l  #0,-(a7)                ;magic number for SymFindFirst()
        move.l  JUMP_TABLE,a0           ;Prepare SymFindFirst()
        move.l  SymFindFirst(a0),a0
        jsr     (a0)                    ;call SymFindFirst()
        add.l   #6,a7                   ;restore stack
CleanUpSysLoop:
        move.l  JUMP_TABLE,a0           ;Prepare SymFindNext()
        move.l  SymFindNext(a0),a0
        jsr     (a0)                    ;call SymFindNext()
        cmp.l   #0,a0                   ;test return value
        beq     CleanUpSysEnd           ;no more files
        move.l  8(a0),d0                ;get attributes
        and.l   #$00000280,d0           ;isolate folder and archive bit
        tst.l   d0                      ;if set
        bne     CleanUpSysLoop          ;then process next
        move.w  12(a0),d0               ;get file handle
        _LDEREF d0,a0
        add.l   #2,a0
        move.l  4(a0),d1
        and.l   #$ffffff00,d1           ;get the first 3 characters
        cmp.l   #$36386b00,d1           ;compare with '68k'
        bne     CleanUpSysLoop          ;Wrong file format
        move.w  d0,-(a7)                ;save handle
        tst.w   8(a0)                   ;test RelocCount
        beq     DontClean               ;Skip if not clean
        move.w  #1,8(a0)                ;Set reloc count to 1
        move.l  a0,a6                   ;let a6 point to program
        bsr     UnrelocFromPtr          ;unrelocate
DontClean:
        move.w  (a7)+,d0                ;restore handle
        _UNLOCK d0                      ;unlock file
        bra     CleanUpSysLoop          ;loop
CleanUpSysEnd:
        rts
        
;*****************************************************************************
;* Gets the address of a file                                               **
;*****************************************************************************
GetFileAddress: ;Returns pointer to file in a0. a0 -> filename.
                ;the handle is in d0
        bsr     FindFile
        tst.w   d0                      ;Test if file found
        beq     GetFileAddressFailure   ;fail if not
        _LDEREF d0,a0                   ;get pointer to file
        and.l   #$200,d1                ;Test if file in archive
        beq     GetFileAddressSuccess   ;Succeed if not
        move.l  6(a0),d1                ;Get file ID
        cmp.l   #$36386b41,d1           ;compare with 68kA
        beq     GetFileAddressSuccess   ;succeed if they match
GetFileAddressFailure:
        move.l  #0,d0                   ;clear handle
        move.l  #0,a0                   ;clear pointer
        rts                             ;return
GetFileAddressSuccess:
        rts                             ;return

;*****************************************************************************
;* Finds a file                                                             **
;* a0 points to filename. Returns handle in d0 and attribs in d1.           **
;*****************************************************************************
FindFile:
        movem.l d2-d7/a1-a6,-(a7)       ;save all regs
        move.l  a0,-(a7)                ;save file name
        move.w  #$0A,-(a7)              ;magic number for SymFindFirst()
        move.l  #0,-(a7)                ;magic number for SymFindFirst()
        move.l  JUMP_TABLE,a0           ;Prepare SymFindFirst()
        move.l  SymFindFirst(a0),a0
        jsr     (a0)                    ;call SymFindFirst()
        add.l   #6,a7                   ;restore stack
        move.l  (a7),a1                 ;let a1 point to filename
        bsr     CompareSymbol           ;compare file names
        tst.l   d0                      ;test result
        bne     FindFileFound           ;found file
FindFileLoop:
        move.l  JUMP_TABLE,a0           ;Prepare SymFindNext()
        move.l  SymFindNext(a0),a0
        jsr     (a0)                    ;call SymFindNext()
        cmp.l   #0,a0                   ;test return value
        beq     FindFileNotFound        ;was last file
        move.l  (a7),a1                 ;let a1 point to filename
        bsr     CompareSymbol           ;compare file names
        tst.l   d0                      ;test result
        bne     FindFileFound           ;found file
        bra     FindFileLoop            ;loop
FindFileFound:
        clr.l   d0                      ;clear d0
        move.w  12(a0),d0               ;put it's handle into d0
        move.l  8(a0),d1                ;put it's attribs into d1
        move.l  d1,d2                   ;copy attribs to d2
        and.l   #$00000080,d2           ;check if folder
        tst.l   d2                      ;if symbol is folder
        bne     FindFileLoop            ;precess next
        add.l   #4,a7                   ;restore stack
        movem.l (a7)+,d2-d7/a1-a6
        rts                             ;return
FindFileNotFound:
        clr.l   d0
        clr.l   d1
        add.l   #4,a7
        movem.l (a7)+,d2-d7/a1-a6
        rts

;*****************************************************************************
;* Compares the symbol at a0 with the symbol name at a1                     **
;* Returns the result in d0                                                 **
;*****************************************************************************
CompareSymbol:
        move.l  (a1),d1
        move.l  4(a1),d2
        cmp.l   (a0),d1
        bne     CompareSymbolNo
        cmp.l   4(a0),d2
        bne     CompareSymbolNo
        move.l  #$FFFFFFFF,d0
        rts
CompareSymbolNo:
        clr.l   d0
        rts

;*****************************************************************************
;* LexOS data area                                                          **
;*****************************************************************************
                        even

RegisterDump:
DRegs                   dc.l    0,0,0,0,0,0,0,0
ARegs                   dc.l    0,0,0,0,0,0,0,0
PCounter                dc.l    0
StatusReg               dc.w    0

OldIntVec               dc.l    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
                        even

        include "krnlmsg.inc"           ;Localized kernel messages

;* Hex look-up table
HexLUT                  dc.b    "0123456789ABCDEF"

;*****************************************************************************
;* RAM addresses                                                            **
;*****************************************************************************
                        even
RAMTable:

_LexOSEnd:


        end
